#include <cmath>
#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>

#include "Shader.h"
#include "ShaderProgram.h"
#include "FrameTimer.h"
#include "VertexBuffer.h"
#include "MeshLoader.h"
#include "Texture.h"
#include "FrameBuffer.h"

#define check_inside // not correct for pyramid

namespace Math {
	void MatVecMult(GLfloat *M, GLfloat *v, GLfloat *dest)
	{
		dest[0] = M[0]*v[0] + M[4]*v[1] + M[8]*v[2];
		dest[1] = M[1]*v[0] + M[5]*v[1] + M[9]*v[2];
		dest[2] = M[2]*v[0] + M[6]*v[1] +M[10]*v[2];
	}

	void Cross(GLfloat *v1, GLfloat *v2, GLfloat *dest)
	{
		dest[0] = v1[1] * v2[2] - v1[2] * v2[1];
		dest[1] = v1[2] * v2[0] - v1[0] * v2[2];
		dest[2] = v1[0] * v2[1] - v1[1] * v2[0];
	}

	inline GLfloat Dot(GLfloat *v1, GLfloat *v2)
	{
		return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
	}

	inline GLfloat Length(GLfloat *v)
	{
		return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
	}

	void Normalize(GLfloat *v)
	{
		GLfloat length = Length(v);

		v[0] /= length;
		v[1] /= length;
		v[2] /= length;
	}

	inline void Negate(GLfloat *v)
	{
		v[0] = -v[0];
		v[1] = -v[1];
		v[2] = -v[2];
	}

	void FillMatrix(GLfloat *M, GLfloat *v1, GLfloat *v2, GLfloat *v3)
	{
		M[0]  = v1[0]; M[1]  = v1[1]; M[2]  = v1[2]; M[3]  = 0.0f;
		M[4]  = v2[0]; M[5]  = v2[1]; M[6]  = v2[2]; M[7]  = 0.0f;
		M[8]  =-v3[0]; M[9]  =-v3[1]; M[10] =-v3[2]; M[11] = 0.0f;
		M[12] = 0.0f; M[13]  = 0.0f;  M[14] = 0.0f;  M[15] = 1.0f;
	}
}


int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	int width = 1280;
	int height = 720;
	int flags = SDL_OPENGL;
	Uint8 *keystate;

	// Camera variables
	bool grab_mouse = true;
	GLfloat defaultCameraInvPos[3] = {0.0f, -4.0f, 0.0f};
	GLfloat cameraInvPos[3]; memcpy(cameraInvPos, defaultCameraInvPos, sizeof(GLfloat)*3);
	GLfloat cameraInvRot[3] = {0.0f, 0.0f, 0.0f};
	GLfloat dx, dz;
	const float fastMove_step = 0.005f;
	const float slowMove_step = 0.001f;
	float move_step = fastMove_step;
	GLfloat modelview[16];

	// Parallax variables
	const GLfloat defaultScale = 0.015f;
	GLfloat scale = defaultScale;
	const GLfloat scale_step = 0.001f;
	const GLfloat defaultBias = -0.0125f;
	GLfloat bias = defaultBias;
	const GLfloat bias_step = 0.001f;

	// SSAO variables
	const GLfloat defaultSsaoIntensity = 1.0f;
	GLfloat ssaoIntensity = defaultSsaoIntensity;
	const GLfloat ssaoIntensity_step = 0.1f;

	// Light variables
	const GLfloat defaultRadius = 1.5f;
	const GLfloat radius_step = 0.5f;
	const GLfloat defaultSpotAngleFrac = 0.2f;
	const GLfloat spotAngleFrac_step = 0.1f;
	GLfloat defaultLightDir[3] = {-0.85f, -1.0f, 0.0f};
	GLfloat defaultPointPos[3] = {0.0f, 0.0f, -10.0f};
	GLfloat defaultSpotPos[3] = {0.0f, 5.0f, -9.0f};
	GLfloat defaultSpotDir[3] = {0.0f,-1.0f, 0.0f};

	GLfloat lightDir[3]; memcpy(lightDir, defaultLightDir, sizeof(GLfloat)*3);
	GLfloat tLightDir[3];
	GLfloat pointRadius = defaultRadius;
	GLfloat pointPos[3]; memcpy(pointPos, defaultPointPos, sizeof(GLfloat)*3);
	GLfloat spotRadius = defaultRadius*2.0f;
	GLfloat spotPos[3]; memcpy(spotPos, defaultSpotPos, sizeof(GLfloat)*3);
	GLfloat spotDir[3]; memcpy(spotDir, defaultSpotDir, sizeof(GLfloat)*3);
	GLfloat spotAngleFrac = defaultSpotAngleFrac; // a fraction of Pi
	GLfloat tPointPos[3], tSpotPos[3];
	GLfloat tSpotDir[3], tSpotDirUp[3], tSpotDirRight[3];
	GLfloat spotMatrix[16];

	GLfloat *keyVec = lightDir; // current keyboard controlled light vector
	GLfloat *keyRad; // current keyboard controlled light radius

	// DoF variables
	const float defaultDistance = 9.0f;
	float focalDistance = defaultDistance;
	const float defaultRange = 5.0f;
	float focalRange = defaultRange;

	GLfloat lRot = 0.0f;
	GLfloat zFar = 60.0; // for SSAO LoD
	enum editMode {PARALLAX, SSAO, DIRECTIONAL, POINT, SPOT, SPOT_DIR, DOF} currentMode = DIRECTIONAL;
	bool quit = false;
	bool with_lmove = false;
	bool with_ambient = true;
	bool with_directional = true;
	bool with_pointlight = true;
	bool with_spotlight = true;
	bool with_blend = true;
	bool with_depthComparison = true;
	bool with_ssao = true;
	bool with_ssaoBlur = true;
	bool with_dof = true;
	bool with_dofBlur = false;

	FrameTimer t(5);


// ----- SDL init ---------------
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);

	
// ----- OpenGL attribute setting via SDL ---------------
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 0);
	SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, 1);

	
// ----- Setting up the screen surface ---------------
	if ((screen = SDL_SetVideoMode(width, height, 0, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("GLSL_deferred by Encelo", NULL);


// ----- OpenGL init ---------------
	glBlendFunc(GL_ONE, GL_ONE);

	// Shaders
	ShaderProgram gbuffer_sh("shaders/gbuffer.vert", "shaders/gbuffer.frag");
	ShaderProgram gbuffer_normal_sh("shaders/gbuffer_parallax.vert", "shaders/gbuffer_normal.frag");
	ShaderProgram gbuffer_parallax_sh("shaders/gbuffer_parallax.vert", "shaders/gbuffer_parallax.frag");
	ShaderProgram gbuffer_iparallax_sh("shaders/gbuffer_parallax.vert", "shaders/gbuffer_iparallax.frag");
	ShaderProgram ssao_sh("shaders/simple.vert", "shaders/ssao.frag");
	ShaderProgram ambient_sh("shaders/simple.vert", "shaders/ambient.frag");
	ShaderProgram directional_sh("shaders/simple.vert", "shaders/directional.frag");
	ShaderProgram point_sh("shaders/pointlight.vert", "shaders/pointlight.frag");
	ShaderProgram spot_sh("shaders/pointlight.vert", "shaders/spotlight.frag");
	ShaderProgram simple_sh("shaders/simple.vert", "shaders/simple.frag");
	ShaderProgram texmix_sh("shaders/simple.vert", "shaders/texmix.frag");
	gbuffer_normal_sh.Uniform("Tex0", 0);
	gbuffer_normal_sh.Uniform("Tex1", 1);
	gbuffer_parallax_sh.Uniform("Scale", scale);
	gbuffer_parallax_sh.Uniform("Bias", bias);
	gbuffer_parallax_sh.Uniform("Tex0", 0);
	gbuffer_parallax_sh.Uniform("Tex1", 1);
	gbuffer_iparallax_sh.Uniform("Scale", scale);
	gbuffer_iparallax_sh.Uniform("Bias", bias);
	gbuffer_iparallax_sh.Uniform("Tex0", 0);
	gbuffer_iparallax_sh.Uniform("Tex1", 1);
	ssao_sh.Uniform("screenSize", width*0.5f, height*0.5f);
	ssao_sh.Uniform("zFar", zFar);
	ssao_sh.Uniform("SampleRad", 1.0f);
	ssao_sh.Uniform("Scale", 1.0f);
	ssao_sh.Uniform("Bias", 0.0f);
	ssao_sh.Uniform("Tex0", 0);
	ssao_sh.Uniform("Tex1", 1);
	ssao_sh.Uniform("Tex2", 2);
	ambient_sh.Uniform("Tex0", 0);
	ambient_sh.Uniform("Tex1", 1);
	ambient_sh.Uniform("lightColor", 0.2f, 0.2f, 0.2f, 1.0f);
	directional_sh.Uniform("lightColor", 1.0f, 1.0f, 1.0f, 1.0f);
	directional_sh.Uniform("Tex0", 0);
	directional_sh.Uniform("Tex1", 1);
	directional_sh.Uniform("Tex2", 2);
	point_sh.Uniform("lightColor", 1.0f, 0.0f, 0.0f, 1.0f);
	point_sh.Uniform("lightIntensity", 1.0f);
	point_sh.Uniform("Tex0", 0);
	point_sh.Uniform("Tex1", 1);
	point_sh.Uniform("Tex2", 2);
	spot_sh.Uniform("lightColor", 0.0f, 0.0f, 1.0f, 1.0f);
	spot_sh.Uniform("lightIntensity", 1.0f);
	spot_sh.Uniform("spotCosCutoff", cosf(spotAngleFrac * M_PI));
	spot_sh.Uniform("Tex0", 0);
	spot_sh.Uniform("Tex1", 1);
	spot_sh.Uniform("Tex2", 2);
	texmix_sh.Uniform("Tex0", 0);
	texmix_sh.Uniform("Tex1", 1);
	texmix_sh.Uniform("Tex2", 2);
	ShaderProgram *bricks_sh = &gbuffer_iparallax_sh;
	ShaderProgram *last_bricks_sh = bricks_sh; // for DoF when rendering blur
	bool last_with_ssao = with_ssao; // for DoF when rendering blur

	ShaderProgram gaussH_sh("shaders/simple.vert", "shaders/gaussianH.frag");
	ShaderProgram gaussV_sh("shaders/simple.vert", "shaders/gaussianV.frag");

	float weights[9] = {0.0677841f, 0.0954044f, 0.121786f, 0.140999f, 0.148054f, 0.140999f, 0.121786f, 0.0954044f, 0.0677841f};
	float offsetsH[9], offsetsV[9];
	for (int i=0; i<9; i++) {
		offsetsH[i] = (i - 4.0f) / float(width/1.0);
		offsetsV[i] = (i - 4.0f) / float(height/1.0);
	}

	gaussH_sh.Uniform("Weights", 9, weights);
	gaussH_sh.Uniform("Offsets", 9, offsetsH);
	gaussV_sh.Uniform("Weights", 9, weights);
	gaussV_sh.Uniform("Offsets", 9, offsetsV);

	// Textures
	Texture randtex("textures/random_normals.png");
	Texture texr0("textures/bricks2_color.jpg");
	Texture texr1("textures/bricks2_parallax.png");
	Texture texl0("textures/marble.jpg");

	// FBOs
	FrameBuffer gbuffer(width, height);
	gbuffer.AttachRender(GL_DEPTH_COMPONENT24);
	gbuffer.AttachTexture(GL_RGBA8);
	gbuffer.AttachTexture(GL_RGBA8);
	gbuffer.AttachTexture(GL_RGBA16F_ARB);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	FrameBuffer ssao_fbo(width/2, height/2);
	ssao_fbo.AttachTexture(GL_RGBA8);
	FrameBuffer scene_fbo(width, height);
	scene_fbo.AttachTexture(GL_RGBA8);
	scene_fbo.AttachRender(GL_DEPTH_COMPONENT24);

	FrameBuffer *blur_fbo = new FrameBuffer[2];
	for (int i = 0; i < 2; i++) {
		blur_fbo[i].SetSize(int(width/2), int(height/2));
		blur_fbo[i].AttachTexture(GL_RGBA8, GL_LINEAR);
		// clamping to preserve post-pro from blur leaking
		blur_fbo[i].BindTex();
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}

	FrameBuffer::Unbind();

	// Meshes
	MeshLoader ml;
	VertexBuffer room, lucy;
	ml.Load("models/uv_room.ply", &room);
	ml.Load("models/hi_lucy.ply", &lucy);
//	ml.Load("models/uv_lucy.ply", &lucy);

	// VBOs
	VertexBuffer quad;
	GLfloat qverts[] = {-1,-1, 0,  1,-1, 0,  1, 1, 0, -1, 1, 0};
	GLfloat qcoords[] = {0, 0,  1, 0,  1, 1,  0, 1};
	GLuint qind[] = {0, 1, 2,  0, 2, 3};
	quad.SetVertices(4, qverts);
	quad.SetTexCoords(4, qcoords);
	quad.SetIndices(6, qind);


	VertexBuffer cube;
	GLfloat cverts[] = {
			-1.0,-1.0, 1.0,  1.0,-1.0, 1.0,  1.0, 1.0, 1.0, -1.0, 1.0, 1.0,
			-1.0,-1.0,-1.0,  1.0,-1.0,-1.0,  1.0, 1.0,-1.0, -1.0, 1.0,-1.0,
	};
	GLuint cind[] = {
			0, 1, 2,  0, 2, 3, // Front
			3, 2, 6,  3, 6, 7, // Top
			7, 6, 5,  7, 5, 4, // Back
			4, 5, 1,  4, 1, 0, // Bottom
			1, 5, 6,  1, 6, 2, // Right
			4, 0, 3,  4 ,3 ,7  // Left
	};
	cube.SetVertices(8, cverts);
	cube.SetIndices(36, cind);

	VertexBuffer pyramid;
	GLfloat pverts[] = {0.0, 0.0, 0.0, -1.0,-1.0,-1.0,  1.0,-1.0,-1.0,  1.0, 1.0,-1.0, -1.0, 1.0,-1.0};
	GLuint pind[] = {0, 1, 2,  0, 2, 3,  0, 3, 4,  0, 4, 1,  1, 4, 3,  1, 3, 2};
	pyramid.SetVertices(5, pverts);
	pyramid.SetIndices(18, pind);

	t.Reset();


// ----- Event cycle ---------------
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_QUIT:
				quit = true;
				break;
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = true;
					break;
				case SDLK_F1:
					flags = screen->flags; // Save current flags in case toggling fails
					screen = SDL_SetVideoMode(0, 0, 0, screen->flags ^ SDL_FULLSCREEN); // Toggles fullScreen mode
					if(screen == NULL) screen = SDL_SetVideoMode(0, 0, 0, flags); // switch back if fails
					if(screen == NULL) exit(1); // exit if can't switch back
					break;
				case SDLK_TAB:
					grab_mouse = !grab_mouse;
					if (grab_mouse) {
						SDL_ShowCursor(SDL_DISABLE);
						SDL_WarpMouse(int(width*0.5f), int(height*0.5f));
					}
					else {
						SDL_ShowCursor(SDL_ENABLE);
					}
					break;
				case SDLK_l:
					with_lmove = !with_lmove;
					break;
				case SDLK_1:
					bricks_sh = &gbuffer_sh;
					break;
				case SDLK_2:
					bricks_sh = &gbuffer_normal_sh;
					break;
				case SDLK_3:
					currentMode = PARALLAX;
					bricks_sh = &gbuffer_parallax_sh;
					break;
				case SDLK_4:
					currentMode = PARALLAX;
					bricks_sh = &gbuffer_iparallax_sh;
					break;
				case SDLK_5:
					currentMode = SSAO;
					break;
				case SDLK_6:
					currentMode = DIRECTIONAL;
					keyVec = lightDir;
					break;
				case SDLK_7:
					currentMode = POINT;
					keyVec = pointPos;
					keyRad = &pointRadius;
					break;
				case SDLK_8:
					currentMode = SPOT;
					keyVec = spotPos;
					keyRad = &spotRadius;
					break;
				case SDLK_9:
					currentMode = SPOT_DIR;
					keyVec = spotDir;
					break;
				case SDLK_0:
					currentMode = DOF;
					break;
				case SDLK_BACKSPACE:
					cameraInvRot[0] = cameraInvRot[1] = 0.0f;
					memcpy(cameraInvPos, defaultCameraInvPos, sizeof(GLfloat)*3);
					break;
				case SDLK_KP0:
					if (currentMode == PARALLAX) {
						scale = defaultScale;
						bias = defaultBias;
					}
					else if (currentMode == SSAO)
						ssaoIntensity = defaultSsaoIntensity;
					else if (currentMode == DIRECTIONAL)
						memcpy(lightDir, defaultLightDir, sizeof(GLfloat)*3);
					else if (currentMode == POINT) {
						memcpy(pointPos, defaultPointPos, sizeof(GLfloat)*3);
						pointRadius = defaultRadius;
						lRot = 0.0f;
					}
					else if (currentMode == SPOT) {
						memcpy(spotPos, defaultSpotPos, sizeof(GLfloat)*3);
						spotRadius = defaultRadius * 2.0f;
						lRot = 0.0f;
					}
					else if (currentMode == SPOT_DIR) {
						memcpy(spotDir, defaultSpotDir, sizeof(GLfloat)*3);
						spotAngleFrac = defaultSpotAngleFrac;
						lRot = 0.0f;
					}
					else if (currentMode == DOF) {
						focalDistance = defaultDistance;
						focalRange = defaultRange;
					}
					break;
				case SDLK_KP_PERIOD:
					if (currentMode == SSAO)
						with_ssaoBlur = !with_ssaoBlur;
					if (currentMode == POINT || currentMode == SPOT || currentMode == SPOT_DIR)
						with_blend = !with_blend;
					else if (currentMode == DOF) {
						with_dofBlur = !with_dofBlur;
						gbuffer_sh.Uniform("renderBlur", with_dofBlur);
						if (with_dofBlur) {
							last_with_ssao = with_ssao;
							with_ssao = false;
							last_bricks_sh = bricks_sh;
							bricks_sh = &gbuffer_sh;
						}
						else {
							with_ssao = last_with_ssao;
							bricks_sh = last_bricks_sh;
						}
					}
					break;
				case SDLK_KP_MULTIPLY:
					if (currentMode == SSAO)
						with_ambient = !with_ambient;
					if (currentMode == POINT || currentMode == SPOT || currentMode == SPOT_DIR)
						with_depthComparison = !with_depthComparison;
					break;
				case SDLK_KP_ENTER:
					if (currentMode == SSAO) {
						if (with_dofBlur == false)
							with_ssao = !with_ssao;
					}
					else if (currentMode == DIRECTIONAL)
						with_directional = !with_directional;
					else if (currentMode == POINT)
						with_pointlight = !with_pointlight;
					else if (currentMode == SPOT || currentMode == SPOT_DIR)
						with_spotlight = !with_spotlight;
					else if (currentMode == DOF)
						with_dof = !with_dof;
					break;
				case SDLK_KP7:
					if (currentMode == POINT || currentMode == SPOT)
						*keyRad += radius_step;
					else if (currentMode == SPOT_DIR)
						spotAngleFrac += spotAngleFrac_step;
					else if (currentMode == SSAO)
						ssaoIntensity += ssaoIntensity_step;
					break;
				case SDLK_KP1:
					if (currentMode == POINT || currentMode == SPOT) {
						if (*keyRad > radius_step)
							*keyRad -= radius_step;
					}
					else if (currentMode == SPOT_DIR)
						spotAngleFrac -= spotAngleFrac_step;
					else if (currentMode == SSAO) {
						if (ssaoIntensity > ssaoIntensity_step)
							ssaoIntensity -= ssaoIntensity_step;
					}
					break;
				case SDLK_RSHIFT:
				case SDLK_LSHIFT:
					move_step = slowMove_step;
					break;
				default:
					break;
				}
				break;
			case SDL_KEYUP:
				switch (event.key.keysym.sym) {
				case SDLK_RSHIFT:
				case SDLK_LSHIFT:
					move_step = fastMove_step;
					break;
				default:
					break;
				}
				break;
			case SDL_MOUSEMOTION:
				if (!grab_mouse)
					break;
				cameraInvRot[1] += (event.motion.x - width*0.5f)*0.2f;
				cameraInvRot[0] += (event.motion.y - height*0.5f)*0.2f;
				if (grab_mouse)
					SDL_WarpMouse(int(width*0.5f), int(height*0.5f));
				break;
			}
		}
		keystate = SDL_GetKeyState(NULL);
		if (keystate[SDLK_RIGHT] || keystate[SDLK_d])
			dx = -move_step * t.GetInterval();
		else if (keystate[SDLK_LEFT] || keystate[SDLK_a])
			dx = move_step * t.GetInterval();
		if (keystate[SDLK_UP] || keystate[SDLK_w])
			dz = move_step * t.GetInterval();
		else if (keystate[SDLK_DOWN] || keystate[SDLK_s])
			dz = -move_step * t.GetInterval();
		if (keystate[SDLK_SPACE])
			cameraInvPos[1] -= move_step * t.GetInterval();
		else if (keystate[SDLK_LCTRL])
			cameraInvPos[1] += move_step * t.GetInterval();

		if (keystate[SDLK_KP6]) {
			if (currentMode == DIRECTIONAL || currentMode == POINT ||
				currentMode == SPOT || currentMode == SPOT_DIR)
				keyVec[0] +=  move_step * t.GetInterval();
		}
		else if (keystate[SDLK_KP4]) {
			if (currentMode == DIRECTIONAL || currentMode == POINT ||
				currentMode == SPOT || currentMode == SPOT_DIR)
				keyVec[0] -=  move_step * t.GetInterval();
		}
		if (keystate[SDLK_KP8]) {
			if (currentMode == DIRECTIONAL || currentMode == POINT ||
				currentMode == SPOT || currentMode == SPOT_DIR)
				keyVec[2] -=  move_step * t.GetInterval();
			else if (currentMode == PARALLAX)
				scale += scale_step * move_step * t.GetInterval();
			else if (currentMode == DOF)
				focalDistance +=  move_step * t.GetInterval();
		}
		else if (keystate[SDLK_KP5]) {
			if (currentMode == DIRECTIONAL || currentMode == POINT ||
				currentMode == SPOT || currentMode == SPOT_DIR)
				keyVec[2] +=  move_step * t.GetInterval();
			else if (currentMode == PARALLAX) {
				if (scale > scale_step)
					scale -= scale_step * move_step * t.GetInterval();
			}
			else if (currentMode == DOF)
				focalDistance -=  move_step * t.GetInterval();
		}
		if (keystate[SDLK_KP9]) {
			if (currentMode == DIRECTIONAL || currentMode == POINT ||
				currentMode == SPOT || currentMode == SPOT_DIR)
				keyVec[1] +=  move_step * t.GetInterval();
			else if (currentMode == PARALLAX)
				bias += bias_step * move_step * t.GetInterval();
			if (currentMode == DOF)
				focalRange +=  move_step * t.GetInterval();
		}
		else if (keystate[SDLK_KP3]) {
			if (currentMode == DIRECTIONAL || currentMode == POINT ||
				currentMode == SPOT || currentMode == SPOT_DIR)
				keyVec[1] -=  move_step * t.GetInterval();
			else if (currentMode == PARALLAX)
				bias -= bias_step * move_step * t.GetInterval();
			if (currentMode == DOF)
				focalRange -=  move_step * t.GetInterval();
		}

		
// ----- Rendering to the screen ---------------
		t.AddFrame();

		if (with_lmove)
			lRot += 0.001f * t.GetInterval();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f, float(width)/float(height), 0.1f, zFar);

		// Camera
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glRotatef(cameraInvRot[0], 1.0f, 0.0f, 0.0f);
		glRotatef(cameraInvRot[1], 0.0f, 1.0f, 0.0f);
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview);
		cameraInvPos[0] += dx*modelview[0] + dz*modelview[2];
		cameraInvPos[1] += dx*modelview[4] + dz*modelview[6];
		cameraInvPos[2] += dx*modelview[8] + dz*modelview[10];
		dx = dz = 0.0f;
		glTranslatef(cameraInvPos[0], cameraInvPos[1], cameraInvPos[2]); // Translation in camera space
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview); // Current modelview for light position

		// Rendering to the G-Buffer
		gbuffer.BindOutput();
		glEnable(GL_DEPTH_TEST);
		glDepthMask(GL_TRUE); // Only geometry updates the depth buffer
#ifndef check_inside
		glCullFace(GL_BACK);
		glDepthFunc(GL_LESS);
#endif
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// Room not culled, so can be seen all around from the outside
		glDisable(GL_CULL_FACE);
		bricks_sh->Use();
		bricks_sh->Uniform("Scale", scale);
		bricks_sh->Uniform("Bias", bias);
		bricks_sh->Uniform("focalDistance", focalDistance);
		bricks_sh->Uniform("focalRange", focalRange);
		glActiveTexture(GL_TEXTURE0);
		texr0.Bind();
		glActiveTexture(GL_TEXTURE1);
		texr1.Bind();
		glPushMatrix();
		glTranslatef(0.0f, 6.0f, -10.0f);
		glScalef(10.0f, 10.0f, 10.0f);
		room.DrawElements();
		glPopMatrix();
		glEnable(GL_CULL_FACE);

		gbuffer_sh.Use();
		gbuffer_sh.Uniform("focalDistance", focalDistance);
		gbuffer_sh.Uniform("focalRange", focalRange);
		glActiveTexture(GL_TEXTURE0);
		texl0.Bind();
		glPushMatrix();
		glTranslatef(0.0f, 0.6f, -10.0f);
		glScalef(3.0f, 3.0f, 3.0f);
		lucy.DrawElements();
		glPopMatrix();

		glDepthMask(GL_FALSE);
		glDisable(GL_DEPTH_TEST);
	// Screen Space Ambient Occlusion
	if (with_ssao) {
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		glViewport(0, 0, width/2, height/2);
		ssao_fbo.BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		randtex.Bind();
		glActiveTexture(GL_TEXTURE1);
		gbuffer.BindTex(1);
		glActiveTexture(GL_TEXTURE2);
		gbuffer.BindTex(2);
		ssao_sh.Use();
		ssao_sh.Uniform("Intensity", ssaoIntensity);
		quad.DrawElements();

		if (with_ssaoBlur) {
			// Gaussian blur along the X axis
			blur_fbo[0].BindOutput();
			glClear(GL_COLOR_BUFFER_BIT);
			glActiveTexture(GL_TEXTURE0);
			ssao_fbo.BindTex();
			gaussH_sh.Use();
			quad.DrawElements();

			// Gaussian blur along the Y axis
			blur_fbo[1].BindOutput();
			glClear(GL_COLOR_BUFFER_BIT);
			blur_fbo[0].BindTex();
			gaussV_sh.Use();
			quad.DrawElements();
		}

		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glViewport(0, 0, width, height);
	}

		// Deferred shading
	if (with_dof) {
		gbuffer.BlitTo(&scene_fbo, GL_DEPTH_BUFFER_BIT);
		scene_fbo.BindOutput();
	}
	else {
		// For depth buffer the filter must be nearest
		gbuffer.BlitTo(0, GL_DEPTH_BUFFER_BIT, GL_NEAREST);
		FrameBuffer::Unbind();
	}
		glClear(GL_COLOR_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		gbuffer.BindTex(0);

	// Ambient pass
	if (with_ambient) {
		if (with_ssao) {
			glActiveTexture(GL_TEXTURE1);
			if (with_ssaoBlur)
				blur_fbo[1].BindTex(0);
			else
				ssao_fbo.BindTex(0);
		}
		ambient_sh.Use();
		ambient_sh.Uniform("withSSAO", with_ssao);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		quad.DrawElements();
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}

		glActiveTexture(GL_TEXTURE1);
		gbuffer.BindTex(1);
		glActiveTexture(GL_TEXTURE2);
		gbuffer.BindTex(2);

	// Directional light
	if (with_directional) {
		glEnable(GL_BLEND);
		directional_sh.Use();
		Math::MatVecMult(modelview, lightDir, tLightDir);

		directional_sh.Uniform("lightDir", -tLightDir[0], -tLightDir[1], -tLightDir[2]);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		quad.DrawElements();
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glDisable(GL_BLEND);
	}

	if (with_depthComparison)
		glEnable(GL_DEPTH_TEST);
#ifndef check_inside
	glCullFace(GL_FRONT);
	glDepthFunc(GL_GREATER);
#endif

	// Point light
	if (with_pointlight) {
		glPushMatrix();
		point_sh.Use();
		point_sh.Uniform("lightRadius", pointRadius);

		if (with_lmove) {
			pointPos[0] = defaultPointPos[0] + pointRadius * 0.75f * sinf(lRot);
			pointPos[2] = defaultPointPos[2] + pointRadius * 0.75f * cosf(lRot);
		}

		glTranslatef(pointPos[0], pointPos[1], pointPos[2]);

		glGetFloatv(GL_MODELVIEW_MATRIX, modelview);
		tPointPos[0] = modelview[12];
		tPointPos[1] = modelview[13];
		tPointPos[2] = modelview[14];
		point_sh.Uniform("lightPos", tPointPos[0], tPointPos[1], tPointPos[2], 1.0f);

#ifdef check_inside
		// Checking if inside the light volume (which is a cube, not a sphere)
		if (abs(tPointPos[0]) < pointRadius * 1.2f ||
			abs(tPointPos[1]) < pointRadius * 1.2f ||
			abs(tPointPos[2]) < pointRadius * 1.2f) {
			glCullFace(GL_FRONT);
			glDepthFunc(GL_GREATER);
		}
#endif

		point_sh.Uniform("Debug", !with_blend);
		if (with_blend)
			glEnable(GL_BLEND);
		glScalef(pointRadius, pointRadius, pointRadius);
		cube.DrawElements();
#ifdef check_inside
		glDepthFunc(GL_LESS);
		glCullFace(GL_BACK);
#endif
		glDisable(GL_BLEND);
		glPopMatrix();
	}

	// Spot light
	if (with_spotlight) {
		glPushMatrix();
		spot_sh.Use();
		spot_sh.Uniform("lightRadius", spotRadius);
		glTranslatef(spotPos[0], spotPos[1], spotPos[2]);
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview);

		tSpotPos[0] = modelview[12];
		tSpotPos[1] = modelview[13];
		tSpotPos[2] = modelview[14];
		spot_sh.Uniform("lightPos", tSpotPos[0], tSpotPos[1], tSpotPos[2], 1.0f);

		if (with_lmove)
			spotDir[0] = sinf(lRot);
		Math::Normalize(spotDir);
		Math::MatVecMult(modelview, spotDir, tSpotDir);
		spot_sh.Uniform("spotDir", tSpotDir[0], tSpotDir[1], tSpotDir[2]);

		// Arbitrary non parallel up vector
		tSpotDirUp[0] = spotDir[1]; tSpotDirUp[1] = -spotDir[0]; tSpotDirUp[2] = spotDir[2];
		Math::Cross(spotDir, tSpotDirUp, tSpotDirRight);
		Math::Cross(tSpotDirRight, spotDir, tSpotDirUp);
		Math::FillMatrix(spotMatrix, tSpotDirRight, tSpotDirUp, spotDir);

#ifdef check_inside
		// Checking if inside the light volume (not correct, using a cone instead of a pyramid)
		GLfloat tSpotToPos[3];
		memcpy(tSpotToPos, tSpotPos, sizeof(GLfloat)*3);
		Math::Negate(tSpotToPos);
		Math::Normalize(tSpotToPos);
		if (Math::Length(tSpotPos) < spotRadius && acosf(Math::Dot(tSpotToPos, tSpotDir)) < spotAngleFrac*M_PI) {
			glCullFace(GL_FRONT);
			glDepthFunc(GL_GREATER);
		}
#endif

		spot_sh.Uniform("Debug", !with_blend);
		if (with_blend)
			glEnable(GL_BLEND);

		float scaleY = spotRadius * cosf(spotAngleFrac * 0.5f * M_PI);
		float scaleX = 2.0f * spotRadius * sinf(spotAngleFrac * 0.5f * M_PI);
		glScalef(scaleX, scaleY, scaleX);
		glMultMatrixf(spotMatrix);
		pyramid.DrawElements();
#ifdef check_inside
		glDepthFunc(GL_LESS);
		glCullFace(GL_BACK);
#endif
		glDisable(GL_BLEND);
		glPopMatrix();
	}
	glDisable(GL_DEPTH_TEST);

#ifndef check_inside
		glCullFace(GL_BACK);
		glDepthFunc(GL_LESS);
#endif
	// Depth of Field
	if (with_dof) {
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		glActiveTexture(GL_TEXTURE0);
		// Downsampling
		blur_fbo[1].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		scene_fbo.BindTex();
		simple_sh.Use();
		glViewport(0, 0, width/2, height/2);
		quad.DrawElements();

		// Gaussian filtering along the X axis
		blur_fbo[0].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		blur_fbo[1].BindTex();
		gaussH_sh.Use();
		quad.DrawElements();

		// Gaussian filtering along the Y axis
		blur_fbo[1].BindOutput();
		glClear(GL_COLOR_BUFFER_BIT);
		blur_fbo[0].BindTex();
		gaussV_sh.Use();
		quad.DrawElements();

		// Final compositing
		FrameBuffer::Unbind();
		glClear(GL_COLOR_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		scene_fbo.BindTex();
		glActiveTexture(GL_TEXTURE1);
		blur_fbo[1].BindTex();
		glActiveTexture(GL_TEXTURE2);
		gbuffer.BindTex(1);
		texmix_sh.Use();
		glViewport(0, 0, width, height);
		quad.DrawElements();
	}

		SDL_GL_SwapBuffers();
	}


// ----- Quitting ---------------
	SDL_Quit();
	return 0;
}
